#%%
from pyecharts.charts import Page
from pyecharts.globals import ThemeType
import pandas as pd
from pyecharts import options as opts
from pyecharts.charts import Bar, Scatter, Line, Gauge
from scipy.stats import linregress
from .functions import (
    capitalize_first_letter,
    Heating_Degree_days,
    Simple_regeression,
)
import json

# ========================================================================================================
#                                   FUNCTIONS GRAPHS
# ========================================================================================================
season_type = ["heating", "cooling", "heating_cooling"]
months = [
    "Jan",
    "Feb",
    "Mar",
    "Apr",
    "May",
    "Jun",
    "Jul",
    "Aug",
    "Sep",
    "Oct",
    "Nov",
    "Dec",
]


def line_and_bar(
    theme_type: str,
    x_data: list,
    y_bar_plot: list,
    y_name: list,
    y_title: str,
    y1_data_line: list,
    y1_name: list,
    frequency: str,
):
    """
    bar(Y-axis) and line(Y-axis-1) chart together:

    :param **theme_type**: chart theme. Possible options:
            * themeType.LIGHT
            * themeType.DARK
            * themeType.CHALK
            * themeType.ESSOS
            * themeType.INFOGRAPHIC
            * themeType.MACARONS
            * themeType.PURPLE_PASSION
            * themeType.ROMA
            * themeType.ROMANTIC
            * themeType.SHINE
            * themeType.VINTAGE
            * themeType.WALDEN
            * themeType.WESTEROS
            * themeType.WONDERLAND
    :param **x_data**: data for x-axis as str-list
    :param **y_bar_plot**: data for bar_chart. -> list of lists, e.g.[[1,2,3,4],[4,5,6,7]]
    :param **y_name**: name of the data to be visualize as bars. same length of y_bar_plot. -> list of lists, e.g [['Q_H, 'Q_HC]]
    :param **y_title**: title of the y_axes
    :param **y1_data_line**: data for line_chart. -> list of lists, e.g.[[1,2,3,4],[4,5,6,7]]
    :param **y1_name**: name of the data to be visualize as lines. same length of y_bar_plot. -> list of lists, e.g [['Q_H, 'Q_HC]]
    :param **frequency**: data frequency to be used in the graph title. e.g. daily, hourly, etc.

    :return: chart in html format
    """

    x_data = sorted(x_data)
    bar = (
        Bar(init_opts=opts.InitOpts(theme=theme_type))
        .add_xaxis(x_data)
        .extend_axis(
            yaxis=opts.AxisOpts(
                name="Temperature",
                type_="value",
                min_=0,
                max_=45,
                interval=5,
                axislabel_opts=opts.LabelOpts(formatter="{value} °C"),
            )
        )
        .set_global_opts(
            title_opts=opts.TitleOpts(title=f"{frequency} energy need"),
            datazoom_opts=[
                opts.DataZoomOpts(range_start=0, range_end=100),
                opts.DataZoomOpts(type_="inside"),
            ],
            toolbox_opts=opts.ToolboxOpts(
                feature=opts.ToolBoxFeatureOpts(
                    save_as_image=opts.ToolBoxFeatureSaveAsImageOpts(
                        title="Download as Image"
                    ),
                    restore=opts.ToolBoxFeatureRestoreOpts(title="Restore"),
                    data_view=opts.ToolBoxFeatureDataViewOpts(
                        title="View Data", lang=["Data View", "Close", "Refresh"]
                    ),
                    data_zoom=opts.ToolBoxFeatureDataZoomOpts(
                        zoom_title="Zoom In", back_title="Zoom Out"
                    ),
                    magic_type=opts.ToolBoxFeatureMagicTypeOpts(is_show=False),
                )
            ),
            tooltip_opts=opts.TooltipOpts(trigger="axis"),
            yaxis_opts=opts.AxisOpts(
                name=y_title, axislabel_opts=opts.LabelOpts(formatter="{value} W")
            ),
        )
    )

    for i, values in enumerate(y_bar_plot):
        bar.add_yaxis(
            series_name=y_name[i],
            y_axis=values,
            label_opts=opts.LabelOpts(is_show=False),
        )

    # Y-AXIS - 1
    lineChart = Line().add_xaxis(x_data)
    for i, line in enumerate(y1_data_line):
        lineChart.add_yaxis(
            series_name=y1_name[i],
            yaxis_index=1,
            y_axis=line,
            label_opts=opts.LabelOpts(is_show=False),
        )

    bar.height = "600px"
    bar.width = "1400px"

    return bar.overlap(lineChart)


def bar_chart_single(y_name: list, y_data_plot: list, theme_type: str):
    """
    Simple bar chart

    :param y_name: name of the parameters to be plot
    :param y_data_plot: list of data for each single parameter
    :param theme_type: chart theme, possible values are desribed in the parmeters of ``line_and_bar`` function

    :return: simple html bar chart
    
    """
    c = (
        Bar(init_opts=opts.InitOpts(theme=theme_type))
        .add_xaxis(months)
        .set_global_opts(
            title_opts=opts.TitleOpts(title="Monthly consumption"),
            datazoom_opts=[
                opts.DataZoomOpts(range_start=0, range_end=100),
                opts.DataZoomOpts(type_="inside"),
            ],
            toolbox_opts=opts.ToolboxOpts(
                feature=opts.ToolBoxFeatureOpts(
                    save_as_image=opts.ToolBoxFeatureSaveAsImageOpts(
                        title="Download as Image"
                    ),
                    restore=opts.ToolBoxFeatureRestoreOpts(title="Restore"),
                    data_view=opts.ToolBoxFeatureDataViewOpts(
                        title="View Data", lang=["Data View", "Close", "Refresh"]
                    ),
                    data_zoom=opts.ToolBoxFeatureDataZoomOpts(
                        zoom_title="Zoom In", back_title="Zoom Out"
                    ),
                    magic_type=opts.ToolBoxFeatureMagicTypeOpts(is_show=False),
                )
            ),
            tooltip_opts=opts.TooltipOpts(trigger="axis"),
        )
    )
    for i, values in enumerate(y_data_plot):
        c.add_yaxis(y_name[i], values)
    c.height = "600px"
    c.width = "1200px"

    return c


def energy_gauge_chart(
    name_series: str,
    value: float,
    unit: str,
    maxLimit: float = 500,
    title_graph: str = "energy need",
):
    """
    Visualize energy need as gauge chart with different levels of performance,
    according to national energy labels

    :param name_series: name of the variable to be shown
    :param value: variable value to be visualized
    :param unit_: unit of measurement
    :param maxLimit: max limit of energy need, according to national rank. e.g 500 kWh/m2 (last category F)
    :param title_graph: title of the graph. Default: "energy need"

    :return: gauge chart in html format
    """
    value_norm = value
    c = (
        Gauge()
        .add(
            series_name=name_series,
            data_pair=[(f"{unit}", value_norm)],
            axisline_opts=opts.AxisLineOpts(
                linestyle_opts=opts.LineStyleOpts(
                    color=[
                        (0.1, "#024b24"),
                        (0.2, "#00602b"),
                        (0.3, "#1eab39"),
                        (0.4, "#bfd804"),
                        (0.5, "#fff500"),
                        (0.6, "#f7ad02"),
                        (0.7, "#e57818"),
                        (1, "#c5271f"),
                    ],
                    width=30,
                )
            ),
            detail_label_opts=opts.LabelOpts(formatter="{value}"),
        )
        .set_global_opts(
            title_opts=opts.TitleOpts(title=title_graph),
            legend_opts=opts.LegendOpts(is_show=False),
        )
    )

    return c


def Scatter_with_regression(
    x_data: list,
    y_data: list,
    colorPoints: list,
    colorRegression: list,
    chart_title: str,
    seriesName: list,
    name_y_axes: str = "",
    graphName: str = "regressionScatter.html",
    subtitle_eq: str = "",
):
    """
    scatter plot with regression line

    :param x_data: list of data
    :param y_data: list of lists. IE [[1,2,3],[4,5,6]]
    :param colorPoints: list  in which length is equal to number of lists in y_data. I.E. ['red','yellow','green']
    :param colorRegression: list of color for the regression lines
    :param seriesName: list of name for each regression
    :param graphName: Name of the graph
    :param subtitle_eq: equation of regression line

    :return: scatter plot with regression line in html format

    """
    # Create a line chart for the regression line
    line_chart = (
        Line()
        .add_xaxis(x_data)
        .set_series_opts(label_opts=opts.LabelOpts(is_show=False))
    )
    scatter_chart = (
        Scatter()
        .add_xaxis(xaxis_data=x_data)
        .set_series_opts(label_opts=opts.LabelOpts(is_show=False))
        .set_global_opts(
            title_opts=opts.TitleOpts(
                title=chart_title, subtitle=subtitle_eq, pos_left="left", pos_top="top"
            ),
            toolbox_opts=opts.ToolboxOpts(
                feature=opts.ToolBoxFeatureOpts(
                    save_as_image=opts.ToolBoxFeatureSaveAsImageOpts(
                        title="Download as Image"
                    ),
                    restore=opts.ToolBoxFeatureRestoreOpts(title="Restore"),
                    data_view=opts.ToolBoxFeatureDataViewOpts(
                        title="View Data", lang=["Data View", "Close", "Refresh"]
                    ),
                    data_zoom=opts.ToolBoxFeatureDataZoomOpts(
                        zoom_title="Zoom In", back_title="Zoom Out"
                    ),
                    magic_type=opts.ToolBoxFeatureMagicTypeOpts(is_show=False),
                )
            ),
            tooltip_opts=opts.TooltipOpts(trigger="axis"),
            xaxis_opts=opts.AxisOpts(
                type_="value", splitline_opts=opts.SplitLineOpts(is_show=True)
            ),
            yaxis_opts=opts.AxisOpts(
                type_="value",
                axistick_opts=opts.AxisTickOpts(is_show=True),
                splitline_opts=opts.SplitLineOpts(is_show=True),
                name=name_y_axes,
            ),
        )
    )

    r_squared = []
    n = 0
    for data in y_data:
        # calcuation of regression line
        # Perform linear regression using numpy
        slope, intercept, r_value, p_value, std_err = linregress(x_data, data)

        # Calculate R-squared
        r_squared.append(r_value ** 2)

        # Generate predicted y values based on the regression line

        regression_line = [slope * x_value + intercept for x_value in x_data]
        regression_line = [round(element, 2) for element in regression_line]

        line_chart.add_yaxis(
            series_name=seriesName[n],
            y_axis=regression_line,
            is_smooth=True,
            symbol="emptyCircle",
            is_symbol_show=False,
            label_opts=opts.LabelOpts(is_show=True),
            color=colorRegression[n],
        )

        scatter_chart.add_yaxis(
            series_name=seriesName[n],
            y_axis=data,
            symbol_size=10,
            label_opts=opts.LabelOpts(is_show=False),
            color=colorPoints[n],
        )
        n = n + 1

    chart = scatter_chart.overlap(line_chart)
    # Set dimension of graph
    chart.height = "600px"
    chart.width = "1400px"

    return chart


# ===================================================================================================================
#                                               REPORT
# ===================================================================================================================


class Graphs_and_report:
    def __init__(self, df, season: str, building_area:float):
        self.df = df
        self.season = season
        self.area_bui = building_area

        if season in season_type:
            self.season = season
        else:
            raise ValueError(f"Invalid choice for season. Select heating or cooling")

    def single_variable_plot(self, folder_directory: str, name_file: str):
        """
        Create and save in html format the bart chart of heating, cooliing or heating and cooling consumption

        :param folder_directory: directory where the save the html file
        :param name_file: name of the chart file

        :return: html chart
        """
        # filtering dataset only for heating and cooling
        df_HC = self.df.loc[:, ["Q_H", "Q_HC", "Q_C"]]
        # Monthly resample
        try: 
            dfHC_monthly = df_HC.resample("ME").sum()
        except:
            dfHC_monthly = df_HC.resample("M").sum()
        if self.season == "heating":
            y_data = [[value / 1000 for value in dfHC_monthly.loc[:, "Q_H"].tolist()]]
            y_name = "Heating consumption"
            theme_type = ThemeType.ESSOS

        elif self.season == "cooling":
            y_data = [[value / 1000 for value in dfHC_monthly.loc[:, "Q_C"].tolist()]]
            y_name = "Cooling Consumption"
            theme_type = ThemeType.WALDEN

        else:
            y_data = [
                [value / 1000 for value in dfHC_monthly.loc[:, "Q_H"].tolist()],
                [value / 1000 for value in dfHC_monthly.loc[:, "Q_C"].tolist()],
            ]
            y_name = ["Heating", "Cooling"]
            theme_type = ThemeType.ROMA

        Chart = bar_chart_single(y_name, y_data, theme_type)

        file_path = "{}/{}.html".format(folder_directory, name_file)
        Chart.render(file_path)

        return Chart

    def variables_plot(
        self,
        month_selected: bool = False,
        _month: int = 1,
        _frequency: str = "hourly",
        energy_var: str = "heating",
        folder_directory: str = "",
        name_file: str = "energy_profile",
    ):
        """
        Visualize data as a combination of barchart and linechart
        Example:
            barchart: for energy consuption
            linechart: for temperature

        :param month_selected: pecify whether you want to view all months or a single month. True or  False
        :param _month: select specific month from 1 to 12
        :param _frequency: data frequency. Values: ["yearly", "monthly", "daily", "hourly"]. Deafult: 'hourly'
        :param energy_var: energy period to be shown. A value equal to: 'heating', 'cooling', 'heating_and_cooling'. Default: 'heating'
        :param folder_directory: directory where the save the html file
        :param name_file: name of the chart file. default: energy_profile

        :return: line and bar chart in html
        """

        # ===============================================================
        # Dataset
        df_HC = self.df.sort_index(axis=0)
        # ===============================================================
        # Filtering by month
        # Filter by month if requested
        if isinstance(month_selected, bool):
            if month_selected:
                if not 1 <= _month <= 12:
                    raise ValueError("Month must be within the range 1 to 12")
                else:
                    df_HC = df_HC[df_HC.index.month == _month]
        elif month_selected is not None:
            raise ValueError("Value must be either True or False")
        # ===============================================================
        # Filtering by frequency
        df_HC_energy = df_HC.loc[:, ["Q_H", "Q_HC", "Q_C"]]
        df_HC_temp = df_HC.loc[:, ["T_op", "T_ext"]]
        
        try: 
            frequency_mapping = {
                "yearly": "YE",
                "monthly": "ME",
                "daily": "D",
                "hourly": "h",
            }
            # Get the resampling frequency code
            freq = frequency_mapping.get(_frequency, None)
        except:
            frequency_mapping = {
                "yearly": "Y",
                "monthly": "M",
                "daily": "D",
                "hourly": "h",
            }
            # Get the resampling frequency code
            freq = frequency_mapping.get(_frequency, None)

        if freq:
            # Resample energy and temperature data
            df_HC_energy_resampled = df_HC_energy.resample(freq).sum()
            df_HC_temp_resampled = df_HC_temp.resample(freq).mean()
            # Concatenate resampled data along columns
            df_HC = pd.concat([df_HC_energy_resampled, df_HC_temp_resampled], axis=1)
        else:
            raise ValueError("Invalid frequency specified")

        # ===============================================================
        # Filtering by energy vars
        if energy_var == "heating":
            df_HC_energy_resampled = [df_HC_energy_resampled["Q_H"].to_list()]
            y_name_var = ["Q_H"]
        elif energy_var == "cooling":
            df_HC_energy_resampled = [df_HC_energy_resampled["Q_C"].to_list()]
            y_name_var = ["Q_C"]
        elif energy_var == "heating_and_cooling":
            df_HC_energy_resampled = [
                df_HC_energy_resampled["Q_H"].to_list(),
                df_HC_energy_resampled["Q_C"].to_list(),
            ]
            y_name_var = ["Q_H", "Q_C"]
        else:
            raise ValueError(
                "Invalid energy period selected. Possible choices are 'heating','cooling','heating_and_cooling'"
            )
        # temperature vars
        df_HC_temp_resampled = [
            df_HC_temp_resampled["T_op"].to_list(),
            df_HC_temp_resampled["T_ext"].to_list(),
        ]

        # GENERATE GRAPH
        Chart = line_and_bar(
            theme_type=ThemeType.ROMA,
            x_data=df_HC.index.strftime("%Y-%m-%d %H-%M").to_list(),
            y_bar_plot=df_HC_energy_resampled,
            y_name=y_name_var,
            y_title="Energy",
            y1_data_line=df_HC_temp_resampled,
            y1_name=["T_op", "T_ext"],
            frequency=capitalize_first_letter(_frequency),
        )
        # Chart.render(os.getcwd()+f"/pybuildingenergy/charts/{chart_name}.html")
        file_path = "{}/{}.html".format(folder_directory, name_file)
        # directory_chart = os.getcwd()+f"/charts/{chart_name}.html"
        Chart.render(file_path)
        return Chart

    def annual_charts(
        self,
        folder_directory: str = "",
        name_file: str = "Yearly_eNeed_gauge",
    ):
        """
        visualize result as gauge chart. Useful for yearly energy
        Recommended for displaying annual energy consumption for heating or cooling according to categories defined at the national level.

        :param bui_area: gross building area [m2]
        :param folder_directory: directory where the save the html file
        :param name_file: name of the chart file. default: Yearly_eNeed_gauge

        :return: annual_charts in html
        """
        # Dataset
        df_HC = self.df.sort_index(axis=0)
        df_HC_energy = df_HC.loc[:, ["Q_H", "Q_HC", "Q_C"]].resample("YE").sum()
        value = round(df_HC_energy["Q_H"].values[0] / (1000 * self.area_bui))

        Chart = energy_gauge_chart(
            "Heating",
            value,
            unit="kWh/m2",
            title_graph="Annual energy need for  heating",
        )
        # Plot chart in a single html file
        # Chart.render(os.getcwd()+f"/pybuildingenergy/charts/{chart_name}.html")
        file_path = "{}/{}.html".format(folder_directory, name_file)
        # directory_chart = os.getcwd()+f"/charts/{chart_name}.html"
        Chart.render(file_path)

        return Chart

    def energy_signature(
        self,
        _frequency: str = "daily",
        clean_data: bool = True,
        folder_directory: str = "",
        name_file: str = "energy_signature",
    ):
        """
        Plotting energy signature between heating consumption and HDD
        Data:
        :param _frequency: data frequency. Values: ["monthly", "daily"]
        :param clean_data: remove 0 values from dataset
        :param chart_name: Name of the html file
        :param folder_directory: directory where the save the html file
        :param name_file: name of the chart file. default: energy_signature

        :return: energy signature of the building
        """
        df_HC = self.df.sort_index(axis=0)
        df_HC_energy = df_HC.loc[:, ["Q_H", "Q_HC", "Q_C"]]
        df_HC_temp = df_HC.loc[:, ["T_op", "T_ext"]]
        # FREQUENCY MAPPING
        try: 
            frequency_mapping = {"monthly": "ME", "daily": "D"}
            # Get the resampling frequency code
            freq = frequency_mapping.get(_frequency, None)
        except: 
            frequency_mapping = {"monthly": "M", "daily": "D"}
            # Get the resampling frequency code
            freq = frequency_mapping.get(_frequency, None)

        if freq:
            # Resample energy and temperature data
            df_HC_energy_resampled = df_HC_energy.resample(freq).sum()
            df_HC_temp_resampled = df_HC_temp.resample(freq).mean()
            # Concatenate resampled data along columns
            df_HC = pd.concat([df_HC_energy_resampled, df_HC_temp_resampled], axis=1)
        else:
            raise ValueError("Invalid frequency specified")

        # GETTING HDD
        df_ext_temp_HDD = df_HC_temp["T_ext"].resample("D").mean().to_list()
        HDD = Heating_Degree_days(df_ext_temp_HDD, 18)
        # REMOVE 0 VALUES
        df_HDD_Q = pd.DataFrame(
            {"Q_H": df_HC_energy_resampled["Q_H"].to_list(), "HDD": HDD}
        )
        #  REMOVE 0 VALUES IF SELECTED
        if isinstance(clean_data, bool):
            if clean_data:
                df_HDD_Q = df_HDD_Q[df_HDD_Q["Q_H"] != 0]
            else:
                pass  # No cleaning needed
        else:
            raise ValueError("Value must be either True or False")
        # Equation of regeression
        regress = Simple_regeression(
            df_HDD_Q["HDD"].to_list(), df_HDD_Q["Q_H"].to_list(), "HDD"
        )

        # PLOT
        Chart = Scatter_with_regression(
            chart_title="Heating energy need vs HDD",
            x_data=df_HDD_Q["HDD"].to_list(),
            y_data=[df_HDD_Q["Q_H"].to_list()],
            colorPoints=["red"],
            colorRegression=["green"],
            seriesName=["Q_H"],
            subtitle_eq=f"{regress[1]}, r2:{regress[0]}",
        )

        # Chart.render(os.getcwd()+f"/pybuildingenergy/charts/{chart_name}.html")
        file_path = "{}/{}.html".format(folder_directory, name_file)
        # directory_chart = os.getcwd()+f"/charts/{chart_name}.html"
        Chart.render(file_path)

        return Chart

    #
    def bui_analysis_page(self, folder_directory: str, name_file: str):
        """
        Create a simple report with building performance graphs:

        :param folder_directory: directory where the save the html file
        :param name_file: name of the chart file

        :return: html report
        """

        page = Page(layout=Page.SimplePageLayout)
        page.add(
            self.variables_plot(
                energy_var="heating",
                folder_directory=folder_directory,
                name_file="line_chart",
            ),
            self.variables_plot(
                _frequency="monthly",
                folder_directory=folder_directory,
                name_file="monthly_heating_need",
            ),
            self.variables_plot(
                _frequency="monthly",
                folder_directory=folder_directory,
                energy_var="heating_and_cooling",
                name_file="monthly_heating_cooling_need",
            ),
            self.energy_signature(folder_directory=folder_directory),
            self.annual_charts(folder_directory=folder_directory),
        )

        file_path = "{}/{}.html".format(folder_directory, name_file)
        page.render(file_path)

        print("Report created!")
        return json.dumps({"report": "created"})





'''
Calculation of DHW needs according to the ISO 12831-3
The aim is to evalute and demostrate the validity of the calculation proposed by the ISO.
This work doesn't replace the standard but it should be used along with the EPB standard.

Acknowledgments: The work was developed using the regulations and results obtained from the spreadsheet created by the EPBCenter.

Authour: Daniele Antonucci
'''

# ======================================================================================================
#                               Energy Need curve
# ======================================================================================================

import pandas as pd
import calendar
from datetime import date
from typing import Optional


from pybuildingenergy.source.functions import table_B_3, table_B_4, table_B_5_modified
from pybuildingenergy.global_inputs import WATER_DENSITY, WATER_SPECIFIC_HEAT_CAPACITY

# ================================================================================
#                           CALENDAR RESOLUTION (by nation name)
# ================================================================================

def get_calendar_by_name(nation_name: str):
    """
    Return a Workalendar country instance given a human-readable nation name,
    e.g. 'Italy'. Uses the registry when available; falls back to a small map.
    """
    try:
        from workalendar.registry import registry
        cal_cls = registry.get(nation_name)
        if cal_cls is None:
            raise KeyError(nation_name)
        return cal_cls()
    except Exception:
        # Fallback for common EU countries (extend as needed)
        try:
            from workalendar.europe import Italy, Germany, France, Spain, Austria, Switzerland
            fallback = {
                "Italy": Italy,
                "Germany": Germany,
                "France": France,
                "Spain": Spain,
                "Austria": Austria,
                "Switzerland": Switzerland,
            }
            return fallback[nation_name]()
        except Exception as e:
            raise ValueError(
                f"Cannot resolve calendar for nation '{nation_name}'. "
                "Ensure 'workalendar' is installed and the name is valid."
            ) from e


def generate_daily_calendar(year: int, month: int, country_calendar) -> dict:
    """Generates daily status (Holiday, Working, Non-Working) for a given month/year."""
    daily_calendar = {}
    holidays = [h[0] for h in country_calendar.holidays(year)]
    for day in range(1, 32):
        try:
            current_date = date(year, month, day)
        except ValueError:
            break  # stop when month runs out of days
        if current_date in holidays:
            status = "Holiday"
        elif country_calendar.is_working_day(current_date):
            status = "Working"
        else:
            status = "Non-Working"
        daily_calendar[current_date.strftime("%Y-%m-%d")] = status
    return daily_calendar


def generate_calendar(nation_name: str, year: int) -> pd.DataFrame:
    """
    Create a calendar DataFrame for the full year for a given nation name.
    Returns columns: ['days' (Timestamp), 'values' in {'Working','Non-Working','Holiday'}].
    """
    cal = get_calendar_by_name(nation_name)

    rows = []
    for month in range(1, 13):
        monthly = generate_daily_calendar(year, month, cal)
        for day, val in sorted(monthly.items()):
            rows.append((pd.to_datetime(day), val))

    df = pd.DataFrame(rows, columns=["days", "values"]).dropna(subset=["days"])
    df = df.sort_values("days").drop_duplicates(subset=["days"], keep="last").reset_index(drop=True)

    # Ensure full coverage (if anything missing)
    full_range = pd.date_range(f"{year}-01-01", f"{year}-12-31", freq="D")
    existing = set(df["days"].dt.normalize())
    missing = [d for d in full_range if d.normalize() not in existing]
    if missing:
        def label_for(d):
            return "Working" if cal.is_working_day(d) else "Holiday"
        add_rows = [(d, label_for(d)) for d in missing]
        df = pd.concat([df, pd.DataFrame(add_rows, columns=["days", "values"])], ignore_index=True)
        df = df.sort_values("days").reset_index(drop=True)
    return df



def calc_V_w_day_trough_V_w_p_day(method: Optional[str] = None,
                                  building_area: Optional[float] = None,
                                  building_type: Optional[str] = None,
                                  V_w_p_day: Optional[float] = None):
    '''
    for single family dwellings and apartments dwellings the V_W_p_day is calcualted based on the number of equivalent persons (adults)

    :param mode: two methods can be used: 
        1) 'number_of_person' providing the number of person' 
        2) 'building_area': providing the area of the building 
        3) 'number_of_units': provinding specific units according to the table B_4
    :param method: possible selection: 
        1) 'correlation': using correlation of B.5
        2) 'table': using table of B.5
    :param building_area:  gross building area [m2]
    :param building_type:  type of building, possible choice: 'Single_family_house', 'Attached_house', 'Dwelling'
    :param num_person: number of person inhabiting the dwelling
    :param V_w_f_day: value taken from the dataframe "table_b_5"
    :param V_w_p_day: value of liters of DHW per person taken from table b_5 modified

    '''

    if building_type in {'Single_family_house', 'Attached_house'}:
        if building_area < 30: 
            n_p_eq_max = 1
        elif 30 <=building_area <= 70:
            n_p_eq_max = 1.775-0.01875*(70-building_area)
        else:
            n_p_eq_max = 0.025*building_area

    elif building_type == 'Dwelling':
        if building_area < 10: 
            n_p_eq_max = 1
        elif 10 <=building_area <= 50:
            n_p_eq_max = 1.775-0.01875*(50-building_area)
        else:
            n_p_eq_max = 0.035*building_area
    else: 
        raise ValueError("Error")

    if n_p_eq_max < 1.75:
        np_eq = n_p_eq_max
    else: 
        np_eq = 1.75+0.3*(n_p_eq_max -1.75)

    if  method == 'correlation':
        # For these residential cases and at the level of one dwelling, requirements can be expressd by :
        V_w_nd_ref = max(40.71, (3.26*building_area/np_eq))/1000 #[m3]
    elif method == 'table':
        V_w_nd_ref = V_w_p_day*np_eq/1000
    else:
        raise ValueError("Invalid method. Choose either 'correlation' or 'table'.")

    return V_w_nd_ref



# ======================================================================
'''
Calculation based on usage and the floor area
'''

def get_days(year):
    """Calculate the number of days per month for a given year excluding days from adjoining months represented as '0'.
    
    Args:
    :param: year (int): The year for which to calculate days per month.

    Returns:
        list: A list containing the number of valid days for each month of the specified year.
    """
    days_per_month = []
    for month in range(1, 13):  # Looping through each month (1 to 12)
        # Getting the week list for each month
        month_weeks = calendar.monthcalendar(year, month)
        # Counting days excluding '0' which represents days of the adjoining month
        days_count = sum(day != 0 for week in month_weeks for day in week)
        days_per_month.append(days_count) 

    return days_per_month



def Volume_and_energy_DHW_calculation(
    n_workdays: int,
    n_weekends: int,
    n_holidays: int,
    sum_fractions: pd.DataFrame,
    total_days: int,
    hourly_fractions: pd.DataFrame,
    teta_W_draw: float,
    teta_w_c_ref: float,
    teta_w_h_ref: float,
    teta_W_cold: float,
    mode_calc: str,
    building_type_B3: str,
    building_area: float,
    unit_count: int,
    building_type_B5: str,
    residential_typology: str,
    calculation_method: str,
    year: int,
    country_calendar: pd.DataFrame
    ):
    '''
    Calculate the daily, monthly, and yearly energy and volume needs for Domestic Hot Water (DHW)
    based on the building parameters and usage.

    Param
    ------
    :param: n_workdays (int): Number of workdays in the year.
    :param: n_weekends (int): Number of weekend days in the year.
    :param: n_holidays (int): Number of holidays in the year.
    :param: sum_fractions (DataFrame): Sum of hourly usage fractions.
    :param: total_days (int): Total number of days in the year.
    :param: hourly_fractions (DataFrame): Hourly usage fractions for different day types.
    :param: teta_W_draw: Water temperature of the mixed (cold nad hot) water drawn at the tap
    :param: teta_w_c_ref:
    :param: teta_w_h_ref:
    :param: teta_W_cold: cold water temperature
    :param: mode (str): calculation mode to be used to get the volume and energy for DHW:
        1. 'Area': using the area of the building
        2. 'number_of_units': number of units according to table B4, B5 
        3. 'volume_type_bui': if building is 'Single_family_house', 'Attached_house', 'Dwellings' the calcuation
            is based on the number of equivalent persons (adults)
    :param: building_type_B3 (str): Building usage type as defined in table B3.
    :param: building_area (float): Building area in square meters.
    :param: unit_count (int): Number of units (e.g., beds, rooms).
    :param: building_type_B5 (str): Building type as defined in table B5.
    :param: residential_typology (str): Specific residential typology.
    :param: calculation_method (str): Method used for calculations ('correlation', etc.).
    :param: year (int): Year for which the calculations are performed.
    :param: country_calendar (DataFrame): Calendar with 'values' column providing
        'Working', 'Non-Working', or 'Holiday' labels for each day.

    Return
    -------
        tuple: Contains detailed DHW needs calculations including yearly consumption,
            monthly and yearly volume, and detailed hourly breakdowns.
    '''
    
    if not isinstance(country_calendar, pd.DataFrame) or "values" not in country_calendar.columns:
        raise TypeError("country_calendar must be a DataFrame with a 'values' column.")

    if building_type_B5 in {'Single_family_house', 'Attached_house', 'Dwelling'}:
        selection_B5 = table_B_5_modified.loc[
            table_B_5_modified['type_of_building'] == residential_typology, :
        ]
        if selection_B5.empty:
            raise ValueError(
                f"Residential typology '{residential_typology}' not found in table_B_5_modified."
            )
        if len(selection_B5) > 1:
            raise ValueError(
                f"Multiple entries found for residential typology '{residential_typology}' in table_B_5_modified."
            )
        liters_per_person = float(selection_B5.iloc[0]['liters/person_per_day'])
        V_nd_d_ref = calc_V_w_day_trough_V_w_p_day(
            method = calculation_method, 
            building_type = building_type_B5, 
            building_area=building_area, 
            V_w_p_day = liters_per_person
            )

        V_W_nd_d = V_nd_d_ref*(teta_w_h_ref-teta_w_c_ref)/(teta_W_draw-teta_w_c_ref)
        
    elif building_type_B5 in {
            'Accomodation', 'Health establishment wihtout accomodation', 
            'Health establishment without accomodation',
            'Health establishment without accomodation - without laundry',
            'Catering, 2 meals per day. Traditional cusine',
            'Catering, 2 meals per day. Self service', 
            'Catering, 1 meals per day. Tradional cusine', 
            'Catering, 1 meals per day. Self service',
            'Hotel, 1-star without laundry', 'Hotel, 1-star withlaundry', 
            'Hotel, 2-star without laundry', 'Hotel, 2-star withlaundry', 
            'Hotel, 3-star without laundry', 'Hotel, 3-star withlaundry', 
            'Hotel, 4-star and GC without laundry', 'Hotel, 4-star and GC withlaundry', 
            'Sport_establishment'
        }: 
        
        if mode_calc == 'area':
            selection_B3 =  table_B_3.loc[table_B_3['type_of_usage']==building_type_B3, :]
            if selection_B3.empty:
                raise ValueError(f"Building usage '{building_type_B3}' not found in table_B_3.")
            if len(selection_B3) > 1:
                raise ValueError(f"Multiple entries found for usage '{building_type_B3}' in table_B_3.")
            # Area specific in kWh/m2d
            area_specific = float(selection_B3.iloc[0]['Area specific - Wh/m2d'])/1000
            # Q_W_calculation: daily energy need at reference condition
            Q_W_nd_d_ref = area_specific * building_area
            # Daily Volume need at  delivery temperature
            V_W_nd_d = Q_W_nd_d_ref/(WATER_DENSITY*WATER_SPECIFIC_HEAT_CAPACITY*(teta_W_draw-teta_w_c_ref))
        elif mode_calc == 'number_of_units':
            selection_B3 =  table_B_3.loc[table_B_3['type_of_usage']==building_type_B3, :]
            if selection_B3.empty:
                raise ValueError(f"Building usage '{building_type_B3}' not found in table_B_3.")
            if len(selection_B3) > 1:
                raise ValueError(f"Multiple entries found for usage '{building_type_B3}' in table_B_3.")
            energy_need = float(selection_B3.iloc[0]['Usage dependent'])
            Q_W_nd_d_ref = unit_count * energy_need 
            # Daily Volume need at  delivery temperature
            V_W_nd_d = Q_W_nd_d_ref/(WATER_DENSITY*WATER_SPECIFIC_HEAT_CAPACITY*(teta_W_draw-teta_w_c_ref))
        elif mode_calc == 'volume_type_bui':
            
            selection_B4 = table_B_4.loc[table_B_4['type_of_activity']==building_type_B5,:]
            if selection_B4.empty:
                raise ValueError(f"Activity '{building_type_B5}' not found in table_B_4.")
            if len(selection_B4) > 1:
                raise ValueError(f"Multiple entries found for activity '{building_type_B5}' in table_B_4.")
            # Daily volume'
            V_nd_d_ref = (unit_count*float(selection_B4.iloc[0]['V_W_f_day']))/1000
            # Daily volume need at delivery temperature
            V_W_nd_d = V_nd_d_ref*(teta_w_h_ref-teta_w_c_ref)/(teta_W_draw-teta_w_c_ref)
        else: 
            raise ValueError("select the right calculation mode from 'area', 'number_of_units', 'volume_type_bui'")
    else: 
        raise ValueError("select the building typology according to those defined in the table")

    # Daily energy need
    V_W_nd_d = float(V_W_nd_d)
    Q_W_nd_d = float(V_W_nd_d * WATER_DENSITY * WATER_SPECIFIC_HEAT_CAPACITY * (teta_W_draw-teta_W_cold))
    # Q_W = area_specific * building_area * n_day
    
    # Monthly Energy Need Calculation
    monthly_cons = [ days * Q_W_nd_d for days in get_days(year)]
    # Yearly Energy Need total of monthly energy needs
    yearly_cons = sum(monthly_cons)

    # Monthly Volume needs
    monthly_volume = [ days * V_W_nd_d for days in get_days(year)]
    # Yearly Volume needs
    yearly_volume = sum(monthly_volume)

    # Totally yearly fractions for workdays
    fractions_workday = n_workdays*sum_fractions.T['Workday'].values[0]
    # Totally yearly fractions for weekend
    fractions_weekend = n_weekends*sum_fractions.T['Weekend'].values[0]
    # Totally yearly fractions for holidays
    fractions_holiday = n_holidays*sum_fractions.T['Holiday'].values[0]
    # 
    tot_fractions = fractions_workday + fractions_weekend + fractions_holiday
    # AVerage correction fractor
    fx_avg = total_days / tot_fractions 
    # Hourly needs, corrected fractions of daily needs
    x_q_h_i_coor = hourly_fractions * fx_avg # dataframe with corrected fraction for workdays weekend and holidays
    
    # Hourly need as a volume at teta_draw
    V_W_nd_h_i = x_q_h_i_coor * V_W_nd_d  # dataframe with corrected fraction for workdays weekend and holidays
    # Hourly need as energy
    Q_W_nd_h_i = x_q_h_i_coor * Q_W_nd_d
    # Hourly result
    # check day category (workday, weekend and holidays): 
    daily_cons_volume = [] 
    daily_cons_energy = []
    for i, day_type in country_calendar.iterrows():
        if day_type['values'] == 'Working':
            col = 'Workday'
        elif day_type['values'] == 'Non-Working':
            col = 'Weekend'
        else:
            col = 'Holiday'
        for item_V in V_W_nd_h_i[col].values.tolist():
            daily_cons_volume.append(item_V)
        for item_Q in Q_W_nd_h_i[col].values.tolist():
            daily_cons_energy.append(item_Q)
    
    return yearly_cons, V_W_nd_d, monthly_volume, yearly_volume, Q_W_nd_d, V_W_nd_h_i, daily_cons_volume, daily_cons_energy


import numpy as np
import pandas as pd
from datetime import datetime
import holidays
# try:
#     import holidays  # type: ignore
# except ImportError:  # pragma: no cover - optional dependency
#     holidays = None  # defer error to runtime with better message

import requests
import plotly.express as px
import plotly.graph_objects as go

def get_country_code_from_latlon(lat, lon, default='IT'):
    """
    Perform a reverse geocode (using e.g. OpenCage or other API) to fetch
    the ISO 3166-1 alpha-2 country code for the given lat,lon.
    If it fails, returns default.
    """
    # Example using OpenCage Data API (you’ll need an API key)
    api_key = "YOUR_OPENCAGE_API_KEY"
    url = "https://api.opencagedata.com/geocode/v1/json"
    params = {
        'q': f"{lat},{lon}",
        'key': api_key,
        'no_annotations': 1,
        'language': 'en'
    }
    resp = requests.get(url, params=params)
    if resp.status_code != 200:
        return default
    data = resp.json()
    if not data['results']:
        return default
    components = data['results'][0].get('components', {})
    country_code = components.get('ISO_3166-1_alpha2')
    if country_code:
        return country_code.upper()
    return default

class HourlyProfileGenerator:
    """
    Generate an hourly DataFrame (13 months by default) with working day and holiday flags
    and profile columns for each category: ventilation, heating, cooling,
    occupancy, lighting, appliances.

    Parameters
    ----------
    country : str
        Country code for national holidays (e.g., 'IT', 'US', 'GB', 'FR', 'DE').
    num_months : int
        Number of months to generate (default: 13).
    start_year : int | None
        Starting year; if None, starts from December of the previous year to the current year.
    working_day_profile : np.array(24) | None
        (Retro-compatibilità) profilo 24h usato se non si forniscono profili per categoria.
    holiday_profile : np.array(24) | None
        (Retro-compatibilità) profilo 24h usato se non si forniscono profili per categoria.
    category_profiles : dict | None
        Dictionary with keys in categories:
        {'ventilation','heating','cooling','occupancy','lighting','appliances'}.
        Each value can be:
          - np.array/list of 24 values → used for weekdays and holidays
          - tuple/list of two arrays of 24 → (weekday, holiday)
          - dict with keys 'weekday' and 'holiday' (each array of 24)

        Example:
        category_profiles = {
            'ventilation': {'weekday': v_wd, 'holiday': v_hd},
            'heating': (h_wd, h_hd),
            'lighting': L_24h,               # same for weekdays/holidays
        }
    """

    CATEGORIES = ("ventilation", "heating", "cooling", "occupancy", "lighting", "appliances")

    def __init__(self,
                 country='IT',
                 num_months=13,
                 start_year=None,
                 working_day_profile=None,
                 holiday_profile=None,
                 category_profiles=None):
        self.country = country
        self.num_months = num_months

        # Start year
        current_date = datetime.now()
        self.current_year = current_date.year
        self.start_year = start_year if start_year else self.current_year - 1
        self.start_date = f"{self.start_year}-12-01"

        # Default "storici" (for retro-compatibility)
        default_working = np.array(
            [1,1,1,1,1,1,0.5,0.5,0.5,0.1,0.1,0.1,0.2,0.2,0.2,0.2,0.5,0.5,0.5,0.8,0.8,0.8,1,1],
            dtype=float
        )
        default_holiday = np.array(
            [1,1,1,1,1,1,0.8,0.8,0.8,0.8,0.8,0.8,0.8,0.8,0.8,0.8,0.8,0.8,0.8,0.8,0.8,0.8,1,1],
            dtype=float
        )

        # If passed, use them instead of defaults
        if working_day_profile is not None:
            self._validate_24(working_day_profile, "working_day_profile")
            default_working = np.asarray(working_day_profile, dtype=float)
        if holiday_profile is not None:
            self._validate_24(holiday_profile, "holiday_profile")
            default_holiday = np.asarray(holiday_profile, dtype=float)

        # Build profiles for categories
        self.profiles = self._build_category_profiles(category_profiles, default_working, default_holiday)

        self.df = None
        self.country_holidays = None

    # ------------------------- helpers -------------------------
    @staticmethod
    def _validate_24(arr, name):
        arr = np.asarray(arr)
        if arr.shape != (24,):
            raise ValueError(f"{name}: must have exactly 24 values (shape {arr.shape})")

    def _coerce_pair(self, value, cat_name):
        """
        Normalizes 'value' into a pair (weekday, holiday), each np.array(24).
        Accepts:
          - array/list 24 → (arr, arr)
          - (arr_wd, arr_hd) → two arrays of 24
          - {'weekday': arr_wd, 'holiday': arr_hd}
        """
        if isinstance(value, dict):
            if "weekday" not in value or "holiday" not in value:
                raise ValueError(f"{cat_name}: dict deve contenere 'weekday' e 'holiday'")
            wd, hd = value["weekday"], value["holiday"]
            self._validate_24(wd, f"{cat_name}.weekday")
            self._validate_24(hd, f"{cat_name}.holiday")
            return np.asarray(wd, dtype=float), np.asarray(hd, dtype=float)

        if isinstance(value, (list, tuple)) and len(value) == 2 and (
            np.asarray(value[0]).shape == (24,) and np.asarray(value[1]).shape == (24,)
        ):
            wd, hd = value
            self._validate_24(wd, f"{cat_name}[0]")
            self._validate_24(hd, f"{cat_name}[1]")
            return np.asarray(wd, dtype=float), np.asarray(hd, dtype=float)

        # single array of 24 → copy to both
        self._validate_24(value, f"{cat_name}")
        arr = np.asarray(value, dtype=float)
        return arr, arr

    def _build_category_profiles(self, category_profiles, default_working, default_holiday):
        """
        Restituisce un dict:
          cat -> {'weekday': np.array(24), 'holiday': np.array(24)}
        """
        profiles = {}

        # If no `category_profiles` is provided, use defaults for all categories
        if category_profiles is None:
            for c in self.CATEGORIES:
                profiles[c] = {"weekday": default_working.copy(), "holiday": default_holiday.copy()}
            return profiles

        # Normalize provided profiles + complete missing categories with defaults
        for c in self.CATEGORIES:
            if c in category_profiles:
                wd, hd = self._coerce_pair(category_profiles[c], c)
                profiles[c] = {"weekday": wd, "holiday": hd}
            else:
                profiles[c] = {"weekday": default_working.copy(), "holiday": default_holiday.copy()}

        return profiles

    # ------------------------- public API -------------------------
    def generate(self):
        """Genera il DataFrame orario con le colonne dei profili per categoria."""
        end_date = pd.Timestamp(self.start_date) + pd.DateOffset(months=self.num_months)
        date_range = pd.date_range(start=self.start_date, end=end_date, freq="h", inclusive="left")

        df = pd.DataFrame({"datetime": date_range})
        df["date"] = df["datetime"].dt.date
        df["hour"] = df["datetime"].dt.hour
        df["day_of_week"] = df["datetime"].dt.dayofweek  # 0=Mon ... 6=Sun
        df["day_name"] = df["datetime"].dt.day_name()

        # if holidays is None:
        #     raise ImportError(
        #         "The 'holidays' package is required for HourlyProfileGenerator.generate(). "
        #         "Install it via `pip install holidays`."
        #     )

        # National holidays (including two years of margin)
        years_needed = list(range(self.start_year, self.start_year + 3))
        self.country_holidays = holidays.country_holidays(self.country, years=years_needed)

        df["is_holiday"] = df["date"].apply(lambda x: x in self.country_holidays)
        df["is_weekend"] = df["day_of_week"].isin([5, 6])
        df["is_working_day"] = ~(df["is_holiday"] | df["is_weekend"])
        df["holiday_name"] = df["date"].apply(lambda x: self.country_holidays.get(x, ""))

        # Category profiles
        for cat, pair in self.profiles.items():
            wd = pair["weekday"]
            hd = pair["holiday"]
            # Apply profile based on day type
            df[f"{cat}_profile"] = df.apply(
                lambda row: wd[row["hour"]] if row["is_working_day"] else hd[row["hour"]],
                axis=1
            )

        # Historical 'profile_value' column for retro-compatibility (uses occupancy)
        df["profile_value"] = df["occupancy_profile"]
        df["day_type"] = df["is_working_day"].apply(lambda x: "Working Day" if x else "Holiday/Weekend")

        self.df = df
        return self.df

    def get_summary(self):
        """Print a brief summary."""
        if self.df is None:
            raise ValueError("DataFrame not generated yet. Call generate() first.")
        print(f"Current date: {datetime.now().strftime('%Y-%m-%d')}")
        print(f"Start year: {self.start_year} (previous year)")
        print(f"Country: {self.country}")
        print(f"Date range: {self.df['datetime'].min()} to {self.df['datetime'].max()}")
        print(f"\nTotal hours: {len(self.df)}")
        print(f"Working day hours: {int(self.df['is_working_day'].sum())}")
        print(f"Holiday/Weekend hours: {int((~self.df['is_working_day']).sum())}")
        print("\nAvailable profiles per category:")
        for c in self.CATEGORIES:
            print(f"  - {c}: column '{c}_profile'")


    def plot_annual_profiles(self,
                             categories=None,
                             freq="H",                 # "H" per orario, "D" per media giornaliera
                             include_weekend_shading=True,
                             title="Annual Profiles (Hourly/ Daily)"):
        """
        Crea un grafico Plotly interattivo dei profili sull'intero periodo generato.

        Parametri
        ---------
        categories : list[str] | None
            Quali categorie plottare. Se None, usa tutte quelle disponibili tra:
            ventilation, heating, cooling, occupancy, lighting, appliances,
            e 'internal_gains' se presente come colonna.
        freq : {"H","D"}
            "H" = profilo orario; "D" = media giornaliera (resample).
        include_weekend_shading : bool
            Se True, aggiunge bande verticali per weekend/festivi.
        title : str
            Titolo del grafico.

        Ritorna
        -------
        fig : plotly.graph_objects.Figure
        """
        if self.df is None:
            raise ValueError("DataFrame non generato. Chiama generate() prima di plottare.")

        # Categorie disponibili in base alle colonne presenti
        available = []
        for c in ["ventilation", "heating", "cooling", "occupancy", "lighting", "appliances", "internal_gains"]:
            col = f"{c}_profile"
            if col in self.df.columns:
                available.append(c)

        if categories is None:
            categories = available
        else:
            # tieni solo quelle davvero presenti
            categories = [c for c in categories if c in available]
            if not categories:
                raise ValueError("Nessuna delle categorie richieste è presente nel DataFrame.")

        # Costruisci dataframe lungo per plotly
        cols = [f"{c}_profile" for c in categories]
        plot_df = self.df[["datetime", "is_working_day", "is_weekend", "is_holiday"] + cols].copy()

        if freq.upper() == "D":
            # media giornaliera
            plot_df = (
                plot_df
                .set_index("datetime")
                .resample("D")
                .mean(numeric_only=True)
                .reset_index()
            )
            # dopo il resample le colonne boolean possono diventare float: rigeneriamo flag semplici (weekend/holiday se media>0.5)
            for flag in ["is_working_day", "is_weekend", "is_holiday"]:
                if flag in plot_df.columns:
                    plot_df[flag] = (plot_df[flag] > 0.5).astype(bool)

        # melt
        long_df = plot_df.melt(id_vars=["datetime"], value_vars=cols, var_name="category", value_name="value")
        long_df["category"] = long_df["category"].str.replace("_profile", "", regex=False)

        # line plot
        fig = px.line(
            long_df,
            x="datetime", y="value",
            color="category",
            title=title,
            labels={"value": "profile value (0..1)", "datetime": "Date/Time", "category": "Category"}
        )

        # Opzionale: bande weekend/festivi
        if include_weekend_shading:
            # Usa il df (già orario o giornaliero) per recuperare giorni weekend/festivi
            base = plot_df[["datetime"]].copy()
            # ricava giorno data-only
            base["date"] = base["datetime"].dt.floor("D")
            # prendi un solo record per giorno con le etichette
            day_flags = (self.df[["datetime", "is_weekend", "is_holiday"]]
                         .assign(date=lambda d: d["datetime"].dt.floor("D"))
                         .groupby("date")[["is_weekend", "is_holiday"]].max().reset_index())

            # helper per aggiungere bande
            def _add_bands(mask_col, color_rgba):
                m = day_flags[mask_col].values
                dates = day_flags["date"].values
                for i, is_mask in enumerate(m):
                    if bool(is_mask):
                        start = pd.Timestamp(dates[i])
                        end = start + pd.Timedelta(days=1)
                        fig.add_vrect(x0=start, x1=end,
                                      fillcolor=color_rgba, opacity=0.08, line_width=0)

            # weekend (leggermente grigio), festivi (arancione chiaro)
            _add_bands("is_weekend", "rgba(128,128,128,1)")
            _add_bands("is_holiday", "rgba(255,165,0,1)")

        # layout pulito
        fig.update_layout(
            hovermode="x unified",
            legend_title_text="Category",
            margin=dict(l=40, r=20, t=60, b=40),
        )
        fig.update_yaxes(range=[0, 1], title="Profile value")

        return fig



import numpy as np
import pandas as pd
import pytest
import os
from pathlib import Path


# ==============================================================================
#                           FIXTURES
# ==============================================================================

@pytest.fixture
def building_data():
    """Fixture per i dati dell'edificio"""
    return {
        "building": {
            "name": "test-cy",
            "azimuth_relative_to_true_north": 41.8,
            "latitude": 46.49018685497359,
            "longitude": 11.327028776009655,
            "exposed_perimeter": 40,
            "height": 3,
            "wall_thickness": 0.3,
            "n_floors": 1,
            "building_type_class": "Residential_apartment",
            "adj_zones_present": False,
            "number_adj_zone": 2,
            "net_floor_area": 100,
            "construction_class": "class_i"
        },
        "adjacent_zones": [
            {
                "name": "adj_1",
                "orientation_zone": {"azimuth": 0},
                "area_facade_elements": np.array([20, 60, 30, 30, 50, 50], dtype=object),
                "typology_elements": np.array(['OP', 'OP', 'OP', 'OP', 'GR', 'OP'], dtype=object),
                "transmittance_U_elements": np.array([0.8196721311475411, 0.8196721311475411, 0.8196721311475411, 0.8196721311475411, 0.5156683855612851, 1.162633192818565], dtype=object),
                "orientation_elements": np.array(['NV', 'SV', 'EV', 'WV', 'HOR', 'HOR'], dtype=object),
                'volume': 300,
                'building_type_class': 'Residential_apartment',
                'a_use': 50
            },
            {
                "name": "adj_2",
                "orientation_zone": {"azimuth": 180},
                "area_facade_elements": np.array([20, 60, 30, 30, 50, 50], dtype=object),
                "typology_elements": np.array(['OP', 'OP', 'OP', 'OP', 'GR', 'OP'], dtype=object),
                "transmittance_U_elements": np.array([0.8196721311475411, 0.8196721311475411, 0.8196721311475411, 0.8196721311475411, 0.5156683855612851, 1.162633192818565], dtype=object),
                "orientation_elements": np.array(['NV', 'SV', 'EV', 'WV', 'HOR', 'HOR'], dtype=object),
                'volume': 300,
                'building_type_class': 'Residential_apartment',
                'a_use': 50
            }
        ],
        "building_surface": [
            {
                "name": "Roof surface",
                "type": "opaque",
                "area": 130,
                "sky_view_factor": 1.0,
                "u_value": 2.2,
                "solar_absorptance": 0.4,
                "thermal_capacity": 741500.0,
                "orientation": {"azimuth": 0, "tilt": 0},
                "name_adj_zone": None
            },
            {
                "name": "Opaque north surface",
                "type": "opaque",
                "area": 30,
                "sky_view_factor": 0.5,
                "u_value": 1.4,
                "solar_absorptance": 0.4,
                "thermal_capacity": 1416240.0,
                "orientation": {"azimuth": 0, "tilt": 90},
                "name_adj_zone": "adj_1"
            },
            {
                "name": "Opaque south surface",
                "type": "opaque",
                "area": 30,
                "sky_view_factor": 0.5,
                "u_value": 1.4,
                "solar_absorptance": 0.4,
                "thermal_capacity": 1416240.0,
                "orientation": {"azimuth": 180, "tilt": 90},
                "name_adj_zone": "adj_2"
            },
            {
                "name": "Opaque east surface",
                "type": "opaque",
                "area": 30,
                "sky_view_factor": 0.5,
                "u_value": 1.2,
                "solar_absorptance": 0.6,
                "thermal_capacity": 1416240.0,
                "orientation": {"azimuth": 90, "tilt": 90},
                "name_adj_zone": None
            },
            {
                "name": "Opaque west surface",
                "type": "opaque",
                "area": 30,
                "sky_view_factor": 0.5,
                "u_value": 1.2,
                "solar_absorptance": 0.7,
                "thermal_capacity": 1416240.0,
                "orientation": {"azimuth": 270, "tilt": 90},
                "name_adj_zone": None
            },
            {
                "name": "Slab to ground",
                "type": "opaque",
                "area": 100,
                "sky_view_factor": 0.0,
                "u_value": 1.6,
                "solar_absorptance": 0.6,
                "thermal_capacity": 405801,
                "orientation": {"azimuth": 0, "tilt": 0},
                "name_adj_zone": None
            },
            {
                "name": "Transparent east surface",
                "type": "transparent",
                "area": 4,
                "sky_view_factor": 0.5,
                "u_value": 5,
                "g_value": 0.726,
                "height": 2,
                "width": 1,
                "parapet": 1.1,
                "orientation": {"azimuth": 90, "tilt": 90},
                "shading": False,
                "shading_type": "horizontal_overhang",
                "width_or_distance_of_shading_elements": 0.5,
                "overhang_proprieties": {"width_of_horizontal_overhangs": 1},
                "name_adj_zone": None
            },
            {
                "name": "Transparent west surface",
                "type": "transparent",
                "area": 4,
                "sky_view_factor": 0.5,
                "u_value": 5,
                "g_value": 0.726,
                "height": 2,
                "width": 1,
                "parapet": 1.1,
                "orientation": {"azimuth": 270, "tilt": 90},
                "shading": False,
                "shading_type": "horizontal_overhang",
                "width_or_distance_of_shading_elements": 0.5,
                "overhang_proprieties": {"width_of_horizontal_overhangs": 1},
                "name_adj_zone": None
            }
        ],
        "units": {
            "area": "m²",
            "u_value": "W/m²K",
            "thermal_capacity": "J/kgK",
            "azimuth": "degrees (0=N, 90=E, 180=S, 270=W)",
            "tilt": "degrees (0=horizontal, 90=vertical)",
            "internal_gain": "W/m²",
            "internal_gain_profile": "Normalized to 0-1",
            "HVAC_profile": "0: off, 1: on"
        },
        "building_parameters": {
            "temperature_setpoints": {
                "heating_setpoint": 20.0,
                "heating_setback": 17.0,
                "cooling_setpoint": 26.0,
                "cooling_setback": 30.0,
                "units": "°C"
            },
            "system_capacities": {
                "heating_capacity": 10000000.0,
                "cooling_capacity": 12000000.0,
                "units": "W"
            },
            "airflow_rates": {
                "infiltration_rate": 1.0,
                "units": "ACH (air changes per hour)"
            },
            "internal_gains": [
                {
                    "name": "occupants",
                    "full_load": 4.2,
                    "weekday": [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.5, 0.5, 0.5, 0.1, 0.1, 0.1, 0.1, 0.2, 0.2, 0.2, 0.5, 0.5, 0.5, 0.8, 0.8, 0.8, 1.0, 1.0],
                    "weekend": [1.0, 1.0, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 1.0, 1.0]
                },
                {
                    "name": "appliances",
                    "full_load": 3,
                    "weekday": [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.7, 0.7, 0.5, 0.5, 0.6, 0.6, 0.6, 0.6, 0.5, 0.5, 0.7, 0.7, 0.8, 0.8, 0.8, 0.6, 0.6],
                    "weekend": [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.7, 0.7, 0.5, 0.5, 0.6, 0.6, 0.6, 0.6, 0.5, 0.5, 0.7, 0.7, 0.8, 0.8, 0.8, 0.6, 0.6],
                },
                {
                    "name": "lighting",
                    "full_load": 3,
                    "weekday": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.15, 0.15, 0.15, 0.15, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.15, 0.15],
                    "weekend": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.15, 0.15, 0.15, 0.15, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.15, 0.15],
                }
            ],
            "construction": {
                "wall_thickness": 0.3,
                "thermal_bridges": 2,
                "units": "m (for thickness), W/mK (for thermal bridges)"
            },
            "climate_parameters": {
                "coldest_month": 1,
                "units": "1-12 (January-December)"
            },
            "heating_profile": {
                "weekday": [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                "weekend": [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
            },
            "cooling_profile": {
                "weekday": [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                "weekend": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0]
            },
            "ventilation_profile": {
                "weekday": [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                "weekend": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0]
            }
        }
    }


@pytest.fixture
def hvac_system_config():
    """Fixture per la configurazione del sistema HVAC"""
    return {
        'emitter_type': 'Floor heating 1',
        'nominal_power': 8,
        'emission_efficiency': 90,
        'flow_temp_control_type': 'Type 2 - Based on outdoor temperature',
        'selected_emm_cont_circuit': 0,
        'mixing_valve': True,
        'mixing_valve_delta': 2,
        'heat_losses_recovered': True,
        'distribution_loss_recovery': 90,
        'simplified_approach': 80,
        'distribution_aux_recovery': 80,
        'distribution_aux_power': 30,
        'distribution_loss_coeff': 48,
        'distribution_operation_time': 1,
        'full_load_power': 27,
        'max_monthly_load_factor': 100,
        'tH_gen_i_ON': 1,
        'auxiliary_power_generator': 0,
        'fraction_of_auxiliary_power_generator': 40,
        'generator_circuit': 'independent',
        'gen_flow_temp_control_type': 'Type A - Based on outdoor temperature',
        'gen_outdoor_temp_data': pd.DataFrame({
            "θext_min_gen": [-7],
            "θext_max_gen": [15],
            "θflw_gen_max": [60],
            "θflw_gen_min": [35],
        }, index=["Generator curve"]),
        'speed_control_generator_pump': 'variable',
        'generator_nominal_deltaT': 20,
        'efficiency_model': 'simple',
        'calc_when_QH_positive_only': False,
        'off_compute_mode': 'full',
    }


@pytest.fixture
def output_dir(tmp_path):
    """Fixture per la directory di output temporanea"""
    test_output = tmp_path / "result_test"
    test_output.mkdir()
    return str(test_output)


# ==============================================================================
#                           TESTS
# ==============================================================================

def test_import_package():
    """Test per verificare che il package sia importabile"""
    import pybuildingenergy as pybui
    assert hasattr(pybui, "__version__")


def test_check_heating_system_inputs(hvac_system_config):
    """Test per la validazione degli input del sistema di riscaldamento"""
    import pybuildingenergy as pybui
    
    res = pybui.check_heating_system_inputs(hvac_system_config)
    
    assert "emitter_type" in res
    assert "messages" in res
    assert "config" in res
    assert res["emitter_type"] == "Floor heating 1"


def test_heating_system_calculator(hvac_system_config):
    """Test per il calcolatore del sistema di riscaldamento"""
    import pybuildingenergy as pybui
    
    calc = pybui.HeatingSystemCalculator(hvac_system_config)
    assert calc is not None


@pytest.mark.parametrize("fix", [True, False])
def test_sanitize_and_validate_bui(building_data, fix):
    """Test per la validazione dei dati dell'edificio"""
    import pybuildingenergy as pybui
    
    bui_result, report = pybui.sanitize_and_validate_BUI(building_data, fix=fix)
    
    assert bui_result is not None
    assert isinstance(report, list)
    
    # Verifica che non ci siano errori critici
    errors = [e for e in report if e["level"] == "ERROR"]
    assert len(errors) == 0, f"Errori trovati: {errors}"


@pytest.mark.slow
def test_iso52016_calculation(building_data, output_dir):
    """Test per il calcolo ISO52016 (può richiedere tempo)"""
    import pybuildingenergy as pybui
    
    # Validazione dati
    bui_checked, issues = pybui.sanitize_and_validate_BUI(building_data, fix=True)
    errors = [e for e in issues if e["level"] == "ERROR"]
    
    assert len(errors) == 0, "Errori nella validazione dei dati"
    
    # Esegui calcolo
    hourly_sim, annual_results_df = pybui.ISO52016.Temperature_and_Energy_needs_calculation(
        bui_checked,
        weather_source="pvgis"
    )
    
    # Verifica risultati
    assert hourly_sim is not None
    assert annual_results_df is not None
    assert len(hourly_sim) > 0
    assert len(annual_results_df) > 0
    
    # Salva risultati
    hourly_sim.to_csv(os.path.join(output_dir, "hourly_sim_test.csv"))
    annual_results_df.to_csv(os.path.join(output_dir, "annual_results_test.csv"))
    
    # Verifica che i file siano stati creati
    assert os.path.exists(os.path.join(output_dir, "hourly_sim_test.csv"))
    assert os.path.exists(os.path.join(output_dir, "annual_results_test.csv"))


def test_dhw_calculation():
    """Test per il calcolo del fabbisogno di acqua calda sanitaria"""
    import pybuildingenergy as pybui
    
    # Parametri
    teta_W_draw = 42
    teta_W_cold = 11.2
    teta_w_h_ref = 60
    teta_w_c_ref = 13.5
    
    hourly_fractions = pd.DataFrame({
        "Workday": [0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 10, 10, 20, 10, 10, 10, 10, 5, 0, 0, 0, 0, 0, 0],
        "Weekend": [0, 0, 0, 0, 0, 0, 0, 0, 5, 10, 10, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        "Holiday": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    })
    
    sum_fractions = pd.DataFrame(hourly_fractions.sum())
    sum_fractions.columns = ["fractions"]
    
    # Genera calendario
    calendar_nation = "Italy"
    italy_calendar = pybui.generate_calendar(calendar_nation, 2023)
    
    n_workdays = sum(italy_calendar['values'] == 'Working')
    n_weekends = sum(italy_calendar['values'] == 'Non-Working')
    n_holidays = sum(italy_calendar['values'] == 'Holiday')
    total_days = italy_calendar.count().values[0]
    
    # Calcolo DHW
    dhw_result = pybui.Volume_and_energy_DHW_calculation(
        n_workdays, n_weekends, n_holidays, sum_fractions, total_days, hourly_fractions,
        teta_W_draw,
        teta_w_c_ref,
        teta_w_h_ref,
        teta_W_cold,
        mode_calc='number_of_units',
        building_type_B3='Residential',
        building_area=142,
        unit_count=10,
        building_type_B5='Dwelling',
        residential_typology='residential_building - simple housing - AVG',
        calculation_method='table',
        year=2015,
        country_calendar=italy_calendar
    )
    
    assert dhw_result is not None
    assert len(dhw_result) > 0


# ==============================================================================
#                           MARKERS
# ==============================================================================

# To run only fast tests: pytest -v -m "not slow"
# To run all tests: pytest -v